package com.sun.org.apache.bcel.internal.generic;

import com.sun.org.apache.bcel.internal.classfile.Constant;
import java.util.Map;
import java.util.HashMap;
import java.util.NoSuchElementException;
import java.util.Iterator;
import java.io.DataOutputStream;
import java.io.ByteArrayOutputStream;
import com.sun.org.apache.bcel.internal.Const;
import java.util.ArrayList;
import java.io.IOException;
import com.sun.org.apache.bcel.internal.util.ByteSequence;
import java.util.List;

public class InstructionList implements Iterable<InstructionHandle>
{
    private InstructionHandle start;
    private InstructionHandle end;
    private int length;
    private int[] byte_positions;
    private List<InstructionListObserver> observers;
    
    public InstructionList() {
        this.start = null;
        this.end = null;
        this.length = 0;
    }
    
    public InstructionList(Instruction i) {
        this.start = null;
        this.end = null;
        this.length = 0;
        this.append(i);
    }
    
    public InstructionList(BranchInstruction i) {
        this.start = null;
        this.end = null;
        this.length = 0;
        this.append(i);
    }
    
    public InstructionList(CompoundInstruction c) {
        this.start = null;
        this.end = null;
        this.length = 0;
        this.append(c.getInstructionList());
    }
    
    public boolean isEmpty() {
        return this.start == null;
    }
    
    public static InstructionHandle findHandle(InstructionHandle[] ihs, int[] pos, int count, int target) {
        int l = 0;
        int r = count - 1;
        do {
            int i = (l + r) / 2;
            int j = pos[i];
            if (j == target) {
                return ihs[i];
            }
            if (target < j) {
                r = i - 1;
            }
            else {
                l = i + 1;
            }
        } while (l <= r);
        return null;
    }
    
    public InstructionHandle findHandle(int pos) {
        int[] positions = this.byte_positions;
        InstructionHandle ih = this.start;
        int i = 0;
        while (i < this.length) {
            if (positions[i] == pos) {
                return ih;
            }
            ih = ih.getNext();
            ++i;
        }
        return null;
    }
    
    public InstructionList(byte[] code) {
        this.start = null;
        this.end = null;
        this.length = 0;
        int count = 0;
        InstructionHandle[] ihs;
        int[] pos;
        try (final ByteSequence bytes = new ByteSequence(code)) {
            ihs = new InstructionHandle[code.length];
            pos = new int[code.length];
            while (bytes.available() > 0) {
                int off = bytes.getIndex();
                pos[count] = off;
                Instruction i = Instruction.readInstruction(bytes);
                InstructionHandle ih;
                if ((i instanceof BranchInstruction)) {
                    ih = this.append((BranchInstruction)i);
                }
                else {
                    ih = this.append(i);
                }
                ih.setPosition(off);
                ihs[count] = ih;
                ++count;
            }
        }
        catch (IOException e) {
            throw new ClassGenException(e.toString(), e);
        }
        System.arraycopy(pos, 0, this.byte_positions = new int[count], 0, count);
        for (int j = 0; j < count; ++j) {
            if ((ihs[j] instanceof BranchHandle)) {
                BranchInstruction bi = (BranchInstruction)ihs[j].getInstruction();
                int target = bi.getPosition() + bi.getIndex();
                InstructionHandle ih2 = findHandle(ihs, pos, count, target);
                if (ih2 == null) {
                    throw new ClassGenException("Couldn't find target for branch: " + bi);
                }
                bi.setTarget(ih2);
                if ((bi instanceof Select)) {
                    Select s = (Select)bi;
                    int[] indices = s.getIndices();
                    int k = 0;
                    while (k < indices.length) {
                        target = bi.getPosition() + indices[k];
                        ih2 = findHandle(ihs, pos, count, target);
                        if (ih2 == null) {
                            throw new ClassGenException("Couldn't find target for switch: " + bi);
                        }
                        s.setTarget(k, ih2);
                        ++k;
                    }
                }
            }
        }
    }
    
    public InstructionHandle append(InstructionHandle ih, InstructionList il) {
        if (il == null) {
            throw new ClassGenException("Appending null InstructionList");
        }
        if (il.isEmpty()) {
            return ih;
        }
        InstructionHandle next = ih.getNext();
        InstructionHandle ret = il.start;
        ih.setNext(il.start);
        il.start.setPrev(ih);
        il.end.setNext(next);
        if (next != null) {
            next.setPrev(il.end);
        }
        else {
            this.end = il.end;
        }
        this.length += il.length;
        il.clear();
        return ret;
    }
    
    public InstructionHandle append(Instruction i, InstructionList il) {
        InstructionHandle ih;
        if ((ih = this.findInstruction2(i)) == null) {
            throw new ClassGenException("Instruction " + i + " is not contained in this list.");
        }
        return this.append(ih, il);
    }
    
    public InstructionHandle append(InstructionList il) {
        if (il == null) {
            throw new ClassGenException("Appending null InstructionList");
        }
        if (il.isEmpty()) {
            return null;
        }
        if (this.isEmpty()) {
            this.start = il.start;
            this.end = il.end;
            this.length = il.length;
            il.clear();
            return this.start;
        }
        return this.append(this.end, il);
    }
    
    private void append(InstructionHandle ih) {
        if (this.isEmpty()) {
            this.end = ih;
            (this.start = ih).setNext(ih.setPrev(null));
        }
        else {
            this.end.setNext(ih);
            ih.setPrev(this.end);
            ih.setNext(null);
            this.end = ih;
        }
        ++this.length;
    }
    
    public InstructionHandle append(Instruction i) {
        InstructionHandle ih = InstructionHandle.getInstructionHandle(i);
        this.append(ih);
        return ih;
    }
    
    public BranchHandle append(BranchInstruction i) {
        BranchHandle ih = BranchHandle.getBranchHandle(i);
        this.append(ih);
        return ih;
    }
    
    public InstructionHandle append(Instruction i, Instruction j) {
        return this.append(i, new InstructionList(j));
    }
    
    public InstructionHandle append(Instruction i, CompoundInstruction c) {
        return this.append(i, c.getInstructionList());
    }
    
    public InstructionHandle append(CompoundInstruction c) {
        return this.append(c.getInstructionList());
    }
    
    public InstructionHandle append(InstructionHandle ih, CompoundInstruction c) {
        return this.append(ih, c.getInstructionList());
    }
    
    public InstructionHandle append(InstructionHandle ih, Instruction i) {
        return this.append(ih, new InstructionList(i));
    }
    
    public BranchHandle append(InstructionHandle ih, BranchInstruction i) {
        BranchHandle bh = BranchHandle.getBranchHandle(i);
        InstructionList il = new InstructionList();
        il.append(bh);
        this.append(ih, il);
        return bh;
    }
    
    public InstructionHandle insert(InstructionHandle ih, InstructionList il) {
        if (il == null) {
            throw new ClassGenException("Inserting null InstructionList");
        }
        if (il.isEmpty()) {
            return ih;
        }
        InstructionHandle prev = ih.getPrev();
        InstructionHandle ret = il.start;
        ih.setPrev(il.end);
        il.end.setNext(ih);
        il.start.setPrev(prev);
        if (prev != null) {
            prev.setNext(il.start);
        }
        else {
            this.start = il.start;
        }
        this.length += il.length;
        il.clear();
        return ret;
    }
    
    public InstructionHandle insert(InstructionList il) {
        if (this.isEmpty()) {
            this.append(il);
            return this.start;
        }
        return this.insert(this.start, il);
    }
    
    private void insert(InstructionHandle ih) {
        if (this.isEmpty()) {
            this.end = ih;
            (this.start = ih).setNext(ih.setPrev(null));
        }
        else {
            this.start.setPrev(ih);
            ih.setNext(this.start);
            ih.setPrev(null);
            this.start = ih;
        }
        ++this.length;
    }
    
    public InstructionHandle insert(Instruction i, InstructionList il) {
        InstructionHandle ih;
        if ((ih = this.findInstruction1(i)) == null) {
            throw new ClassGenException("Instruction " + i + " is not contained in this list.");
        }
        return this.insert(ih, il);
    }
    
    public InstructionHandle insert(Instruction i) {
        InstructionHandle ih = InstructionHandle.getInstructionHandle(i);
        this.insert(ih);
        return ih;
    }
    
    public BranchHandle insert(BranchInstruction i) {
        BranchHandle ih = BranchHandle.getBranchHandle(i);
        this.insert(ih);
        return ih;
    }
    
    public InstructionHandle insert(Instruction i, Instruction j) {
        return this.insert(i, new InstructionList(j));
    }
    
    public InstructionHandle insert(Instruction i, CompoundInstruction c) {
        return this.insert(i, c.getInstructionList());
    }
    
    public InstructionHandle insert(CompoundInstruction c) {
        return this.insert(c.getInstructionList());
    }
    
    public InstructionHandle insert(InstructionHandle ih, Instruction i) {
        return this.insert(ih, new InstructionList(i));
    }
    
    public InstructionHandle insert(InstructionHandle ih, CompoundInstruction c) {
        return this.insert(ih, c.getInstructionList());
    }
    
    public BranchHandle insert(InstructionHandle ih, BranchInstruction i) {
        BranchHandle bh = BranchHandle.getBranchHandle(i);
        InstructionList il = new InstructionList();
        il.append(bh);
        this.insert(ih, il);
        return bh;
    }
    
    public void move(InstructionHandle start, InstructionHandle end, InstructionHandle target) {
        if (start == null || end == null) {
            throw new ClassGenException("Invalid null handle: From " + start + " to " + end);
        }
        if (target == start || target == end) {
            throw new ClassGenException("Invalid range: From " + start + " to " + end + " contains target " + target);
        }
        InstructionHandle ih = start;
        while (ih != end.getNext()) {
            if (ih == null) {
                throw new ClassGenException("Invalid range: From " + start + " to " + end);
            }
            if (ih == target) {
                throw new ClassGenException("Invalid range: From " + start + " to " + end + " contains target " + target);
            }
            ih = ih.getNext();
        }
        InstructionHandle prev = start.getPrev();
        InstructionHandle next = end.getNext();
        if (prev != null) {
            prev.setNext(next);
        }
        else {
            this.start = next;
        }
        if (next != null) {
            next.setPrev(prev);
        }
        else {
            this.end = prev;
        }
        start.setPrev(end.setNext(null));
        if (target == null) {
            if (this.start != null) {
                this.start.setPrev(end);
            }
            end.setNext(this.start);
            this.start = start;
        }
        else {
            next = target.getNext();
            target.setNext(start);
            start.setPrev(target);
            end.setNext(next);
            if (next != null) {
                next.setPrev(end);
            }
            else {
                this.end = end;
            }
        }
        return;
    }
    
    public void move(InstructionHandle ih, InstructionHandle target) {
        this.move(ih, ih, target);
    }
    
    private void remove(InstructionHandle prev, InstructionHandle next) throws TargetLostException {
        InstructionHandle first;
        InstructionHandle last;
        if (prev == null && next == null) {
            first = this.start;
            last = this.end;
            InstructionHandle instructionHandle = null;
            this.end = instructionHandle;
            this.start = instructionHandle;
        }
        else {
            if (prev == null) {
                first = this.start;
                this.start = next;
            }
            else {
                first = prev.getNext();
                prev.setNext(next);
            }
            if (next == null) {
                last = this.end;
                this.end = prev;
            }
            else {
                last = next.getPrev();
                next.setPrev(prev);
            }
        }
        first.setPrev(null);
        last.setNext(null);
        List<InstructionHandle> target_vec = new ArrayList();
        for (InstructionHandle ih = first; ih != null; ih = ih.getNext()) {
            ih.getInstruction().dispose();
        }
        StringBuilder buf = new StringBuilder("{ ");
        for (InstructionHandle ih2 = first; ih2 != null; ih2 = next) {
            next = ih2.getNext();
            --this.length;
            if (ih2.hasTargeters()) {
                target_vec.add(ih2);
                buf.append(ih2.toString(true)).append(" ");
                ih2.setNext(ih2.setPrev(null));
            }
            else {
                ih2.dispose();
            }
        }
        buf.append("}");
        if (!target_vec.isEmpty()) {
            InstructionHandle[] targeted = new InstructionHandle[target_vec.size()];
            target_vec.toArray(targeted);
            throw new TargetLostException(targeted, buf.toString());
        }
        return;
    }
    
    public void delete(InstructionHandle ih) throws TargetLostException {
        this.remove(ih.getPrev(), ih.getNext());
    }
    
    public void delete(Instruction i) throws TargetLostException {
        InstructionHandle ih;
        if ((ih = this.findInstruction1(i)) == null) {
            throw new ClassGenException("Instruction " + i + " is not contained in this list.");
        }
        this.delete(ih);
        return;
    }
    
    public void delete(InstructionHandle from, InstructionHandle to) throws TargetLostException {
        this.remove(from.getPrev(), to.getNext());
    }
    
    public void delete(Instruction from, Instruction to) throws TargetLostException {
        InstructionHandle from_ih;
        if ((from_ih = this.findInstruction1(from)) == null) {
            throw new ClassGenException("Instruction " + from + " is not contained in this list.");
        }
        InstructionHandle to_ih;
        if ((to_ih = this.findInstruction2(to)) == null) {
            throw new ClassGenException("Instruction " + to + " is not contained in this list.");
        }
        this.delete(from_ih, to_ih);
        return;
    }
    
    private InstructionHandle findInstruction1(Instruction i) {
        InstructionHandle ih = this.start;
        while (ih != null) {
            if (ih.getInstruction() == i) {
                return ih;
            }
            ih = ih.getNext();
        }
        return null;
    }
    
    private InstructionHandle findInstruction2(Instruction i) {
        InstructionHandle ih = this.end;
        while (ih != null) {
            if (ih.getInstruction() == i) {
                return ih;
            }
            ih = ih.getPrev();
        }
        return null;
    }
    
    public boolean contains(InstructionHandle i) {
        if (i == null) {
            return false;
        }
        InstructionHandle ih = this.start;
        while (ih != null) {
            if (ih == i) {
                return true;
            }
            ih = ih.getNext();
        }
        return false;
    }
    
    public boolean contains(Instruction i) {
        return this.findInstruction1(i) != null;
    }
    
    public void setPositions() {
        this.setPositions(false);
    }
    
    public void setPositions(boolean check) {
        int max_additional_bytes = 0;
        int additional_bytes = 0;
        int index = 0;
        int count = 0;
        int[] pos = new int[this.length];
        if (check) {
            for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
                Instruction i = ih.getInstruction();
                if ((i instanceof BranchInstruction)) {
                    Instruction inst = ((BranchInstruction)i).getTarget().getInstruction();
                    if (!this.contains(inst)) {
                        throw new ClassGenException("Branch target of " + Const.getOpcodeName((int)i.getOpcode()) + ":" + inst + " not in instruction list");
                    }
                    if ((i instanceof Select)) {
                        InstructionHandle[] targets = ((Select)i).getTargets();
                        int length = targets.length;
                        int j = 0;
                        while (j < length) {
                            InstructionHandle target = targets[j];
                            inst = target.getInstruction();
                            if (!this.contains(inst)) {
                                throw new ClassGenException("Branch target of " + Const.getOpcodeName((int)i.getOpcode()) + ":" + inst + " not in instruction list");
                            }
                            ++j;
                        }
                    }
                    if (!(ih instanceof BranchHandle)) {
                        throw new ClassGenException("Branch instruction " + Const.getOpcodeName((int)i.getOpcode()) + ":" + inst + " not contained in BranchHandle.");
                    }
                }
            }
        }
        Instruction i;
        for (InstructionHandle ih = this.start; ih != null; index += i.getLength(), ih = ih.getNext()) {
            i = ih.getInstruction();
            ih.setPosition(index);
            pos[count++] = index;
            switch (i.getOpcode()) {
                case 167:
                case 168: {
                    max_additional_bytes += 2;
                    break;
                }
                case 170:
                case 171: {
                    max_additional_bytes += 3;
                    break;
                }
            }
        }
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            additional_bytes += ih.updatePosition(additional_bytes, max_additional_bytes);
        }
        count = (index = 0);
        for (InstructionHandle ih = this.start; ih != null; pos[count++] = index, index += i.getLength(), ih = ih.getNext()) {
            i = ih.getInstruction();
            ih.setPosition(index);
        }
        if (this.length == count) {
            this.byte_positions = pos;
        }
        else {
            System.arraycopy(pos, 0, this.byte_positions = new int[count], 0, count);
        }
    }
    
    public byte[] getByteCode() {
        this.setPositions();
        ByteArrayOutputStream b = new ByteArrayOutputStream();
        DataOutputStream out = new DataOutputStream(b);
        try {
            for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
                Instruction i = ih.getInstruction();
                i.dump(out);
            }
            out.flush();
        }
        catch (IOException e) {
            System.err.println(e);
            return new byte[0];
        }
        return b.toByteArray();
    }
    
    public Instruction[] getInstructions() {
        List<Instruction> instructions = new ArrayList();
        try (final ByteSequence bytes = new ByteSequence(this.getByteCode())) {
            while (bytes.available() > 0) {
                instructions.add(Instruction.readInstruction(bytes));
            }
        }
        catch (IOException e) {
            throw new ClassGenException(e.toString(), e);
        }
        return (Instruction[])instructions.toArray(new Instruction[instructions.size()]);
    }
    
    @Override
    public String toString() {
        return this.toString(true);
    }
    
    public String toString(boolean verbose) {
        StringBuilder buf = new StringBuilder();
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            buf.append(ih.toString(verbose)).append("\n");
        }
        return buf.toString();
    }
    
    @Override
    public Iterator<InstructionHandle> iterator() {
        return new Iterator<InstructionHandle>() {
            private InstructionHandle ih = InstructionList.this.start;
            
            @Override
            public InstructionHandle next() throws NoSuchElementException {
                if (this.ih == null) {
                    throw new NoSuchElementException();
                }
                InstructionHandle i = this.ih;
                this.ih = this.ih.getNext();
                return i;
            }
            
            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
            
            @Override
            public boolean hasNext() {
                return this.ih != null;
            }
        };
    }
    
    public InstructionHandle[] getInstructionHandles() {
        InstructionHandle[] ihs = new InstructionHandle[this.length];
        InstructionHandle ih = this.start;
        for (int i = 0; i < this.length; ++i) {
            ihs[i] = ih;
            ih = ih.getNext();
        }
        return ihs;
    }
    
    public int[] getInstructionPositions() {
        return this.byte_positions;
    }
    
    public InstructionList copy() {
        Map<InstructionHandle, InstructionHandle> map = new HashMap();
        InstructionList il = new InstructionList();
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            Instruction i = ih.getInstruction();
            Instruction c = i.copy();
            if ((c instanceof BranchInstruction)) {
                map.put(ih, il.append((BranchInstruction)c));
            }
            else {
                map.put(ih, il.append(c));
            }
        }
        for (InstructionHandle ih = this.start, ch = il.start; ih != null; ih = ih.getNext(), ch = ch.getNext()) {
            Instruction j = ih.getInstruction();
            Instruction c2 = ch.getInstruction();
            if ((j instanceof BranchInstruction)) {
                BranchInstruction bi = (BranchInstruction)j;
                BranchInstruction bc = (BranchInstruction)c2;
                InstructionHandle itarget = bi.getTarget();
                bc.setTarget((InstructionHandle)map.get(itarget));
                if ((bi instanceof Select)) {
                    InstructionHandle[] itargets = ((Select)bi).getTargets();
                    InstructionHandle[] ctargets = ((Select)bc).getTargets();
                    for (int k = 0; k < itargets.length; ++k) {
                        ctargets[k] = (InstructionHandle)map.get(itargets[k]);
                    }
                }
            }
        }
        return il;
    }
    
    public void replaceConstantPool(ConstantPoolGen old_cp, ConstantPoolGen new_cp) {
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            Instruction i = ih.getInstruction();
            if ((i instanceof CPInstruction)) {
                CPInstruction ci = (CPInstruction)i;
                Constant c = old_cp.getConstant(ci.getIndex());
                ci.setIndex(new_cp.addConstant(c, old_cp));
            }
        }
    }
    
    private void clear() {
        InstructionHandle instructionHandle = null;
        this.end = instructionHandle;
        this.start = instructionHandle;
        this.length = 0;
    }
    
    public void dispose() {
        for (InstructionHandle ih = this.end; ih != null; ih = ih.getPrev()) {
            ih.dispose();
        }
        this.clear();
    }
    
    public InstructionHandle getStart() {
        return this.start;
    }
    
    public InstructionHandle getEnd() {
        return this.end;
    }
    
    public int getLength() {
        return this.length;
    }
    
    public int size() {
        return this.length;
    }
    
    public void redirectBranches(InstructionHandle old_target, InstructionHandle new_target) {
        for (InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            Instruction i = ih.getInstruction();
            if ((i instanceof BranchInstruction)) {
                BranchInstruction b = (BranchInstruction)i;
                InstructionHandle target = b.getTarget();
                if (target == old_target) {
                    b.setTarget(new_target);
                }
                if ((b instanceof Select)) {
                    InstructionHandle[] targets = ((Select)b).getTargets();
                    for (int j = 0; j < targets.length; ++j) {
                        if (targets[j] == old_target) {
                            ((Select)b).setTarget(j, new_target);
                        }
                    }
                }
            }
        }
    }
    
    public void redirectLocalVariables(LocalVariableGen[] lg, InstructionHandle old_target, InstructionHandle new_target) {
        for (LocalVariableGen element : lg) {
            InstructionHandle start = element.getStart();
            InstructionHandle end = element.getEnd();
            if (start == old_target) {
                element.setStart(new_target);
            }
            if (end == old_target) {
                element.setEnd(new_target);
            }
        }
    }
    
    public void redirectExceptionHandlers(CodeExceptionGen[] exceptions, InstructionHandle old_target, InstructionHandle new_target) {
        for (CodeExceptionGen exception : exceptions) {
            if (exception.getStartPC() == old_target) {
                exception.setStartPC(new_target);
            }
            if (exception.getEndPC() == old_target) {
                exception.setEndPC(new_target);
            }
            if (exception.getHandlerPC() == old_target) {
                exception.setHandlerPC(new_target);
            }
        }
    }
    
    public void addObserver(InstructionListObserver o) {
        if (this.observers == null) {
            this.observers = new ArrayList();
        }
        this.observers.add(o);
    }
    
    public void removeObserver(InstructionListObserver o) {
        if (this.observers != null) {
            this.observers.remove(o);
        }
    }
    
    public void update() {
        if (this.observers != null) {
            for (InstructionListObserver observer : this.observers) {
                observer.notify(this);
            }
        }
    }
}
